#include "httpd.h"

#include "http_log.h"

/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "co_jsinheritance.h"

#define CO_INH_CLASS 0
#define CO_INH_METHOD 1
#define CO_INH_TARGET 2

static const char *query_classes =
"
select distinct class_name
from co_sys_meta_vtable
  ;
";

static const char *query_implementations =
"
select distinct class_name, method_name, target_class_name from co_sys_meta_vtable
 where class_name != target_class_name
 ;
";

// set when server starts
static const char *scripts[2] = { NULL, NULL };
#define SCRIPT_LINE_SIZE 512

static
const char *
cop_buildClassesScript( server_rec *s, pool *p, PGconn *conn )
{
    PGresult *pgres;
    int ntuples;
    int i;
    char *_script;
    const char *script;
    const char *class_name;
    
    pgres = PQexec(conn, query_classes);
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "Querying classes: %s FAILED!", query_classes );
        return NULL;
    }

    ntuples = PQntuples(pgres);
    if (ntuples==0)
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "No classes in vtable!");
	return NULL;
    }

    _script = (char*)ap_palloc( p, ntuples*SCRIPT_LINE_SIZE*2);
    if (!_script) 
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "Allocation of script for %d tuples FAILED!", ntuples);
	return NULL;
    }

    script = _script;
    for (i=0; i<ntuples; i++) {
	class_name = PQgetvalue(pgres, i, CO_INH_CLASS);
	_script += snprintf( _script, SCRIPT_LINE_SIZE, 
	    /* "function %s(arg_area) { this.init = COObject_init; this.init(arg_area,'%s'); };\n",*/
	    "function class_%s(arg_area) { this.strClass='%s'; this.strArea = arg_area; };\n",
	    class_name, class_name
	);
    }

    PQclear(pgres);
    return script;
}

static
const char *
cop_buildImplementationsScript( server_rec *s, pool *p, PGconn *conn )
{
    PGresult *pgres;
    int ntuples;
    int i;
    char *_script;
    const char *script;
    const char *class_name;
    const char *method_name;
    const char *target_name;
    const char *js_method_name;
    
    pgres = PQexec(conn, query_implementations);
    if (PQresultStatus(pgres) != PGRES_TUPLES_OK)
    {
        PQclear(pgres);
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "Querying impl.: %s FAILED!", query_implementations );
        return NULL;
    }

    ntuples = PQntuples(pgres);
    _script = (char*)ap_palloc( p, ntuples*SCRIPT_LINE_SIZE*3);
    if (!_script) 
    {
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "Allocation of script for %d tuples FAILED!", ntuples);
	return NULL;
    }

    script = _script;
    for (i=0; i<ntuples; i++) {
	class_name = PQgetvalue(pgres, i, CO_INH_CLASS);
	method_name = PQgetvalue(pgres, i, CO_INH_METHOD);
	target_name = PQgetvalue(pgres, i, CO_INH_TARGET);
	if (strcmp(method_name, "delete")==0) js_method_name="del";
	else js_method_name = method_name;

	_script += snprintf( _script, SCRIPT_LINE_SIZE, 
	    "class_%s.prototype.RAW_%s = class_%s_RAW_%s;\n"
	    "class_%s.prototype.EXT_%s = class_%s_EXT_%s;\n"
	    "class_%s.prototype.%s = class_%s_%s;\n", 
	    class_name, js_method_name, target_name, method_name,
	    class_name, js_method_name, target_name, method_name,
	    class_name, js_method_name, target_name, method_name
	);
    }

    PQclear(pgres);
    return script;
}

// class.prototype.method = target_method
const char **
cox_buildInheritanceScript( server_rec *s, pool *p, PGconn *conn )
{
    char** scripts = (char**)ap_palloc( p, sizeof(char*)*2);
ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "cox_buildInheritanceScript: allocated scripts" );
    scripts[0] = cop_buildClassesScript( s, p, conn );
    if (scripts[0]==NULL) {
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "scripts[0]==NULL; setting empty" );
	scripts[0] = ap_pstrdup( p, " ");
    }
    scripts[1] = cop_buildImplementationsScript( s, p, conn );
    if (scripts[1]==NULL) {
	ap_log_error(APLOG_MARK, APLOG_DEBUG|APLOG_NOERRNO, s, "scripts[1]==NULL; setting empty" );
	scripts[1] = ap_pstrdup( p, " ");
    }
    return scripts;
}
