/*
 * Copyright (C) by LeftHand sp. z o.o.
 * See http://www.corporateoffice.pl
 */

/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "co_context.h"
#include "co_utils.h"

#include "mod_coffice.h"

#include "co_log.h" /* CO_LOG_LEVELMASK */

/* a few "context methods" */
int
cou_getLogRequestId( co_context* cnx )
{
	return cnx->id_log_request;
}

void
cou_setLogRequestId( co_context* cnx, int id )
{
	cnx->id_log_request = id;
}

int
cou_getSessionId( co_context* cnx )
{
	return cnx->id_session;
}

void
cou_setSessionId( co_context* cnx, int id )
{
	cnx->id_session = id;
}

bool
cou_isEndLogged( co_context* cnx )
{
	return cnx->end_logged;
}

void
cou_onEndLogged( co_context* cnx )
{
	cnx->end_logged = true;
}

void
cou_offEndLogged( co_context* cnx )
{
	cnx->end_logged = false;
}


/* database interface */
static 
PGresult *
cou_execDBQueryCore( co_context* cnx, 
	ExecStatusType ok, const char* query_fmt, va_list args)
{
    char querystr[MAX_STRING_LEN];
	PGconn* conn;
    PGresult *pgres;
	request_rec *r;

	conn = cnx->dbconn;
	r = cnx->request;

	ap_vsnprintf(querystr, sizeof(querystr), query_fmt, args);
    pgres = PQexec(conn, querystr);

	if (PQresultStatus(pgres) != ok)
	{
        PQclear(pgres);
		return NULL;
   }
	return pgres;
}

/* execDBCommand should indicate success or failure */
void
cou_execDBCommand( co_context* cnx, const char* query_fmt, ... )
{
    va_list args;
	PGresult *pgres;

    va_start(args, query_fmt);
    pgres = cou_execDBQueryCore( cnx, PGRES_COMMAND_OK, query_fmt, args);
    va_end(args);
	PQclear(pgres);
}

/* after SELECT we should get some records... */
/* returns NULL on failure */
PGresult *
cou_execDBSelect( co_context* cnx, const char* query_fmt, ... )
{
    va_list args;
	PGresult *pgres;

    va_start(args, query_fmt);
    pgres = cou_execDBQueryCore( cnx, PGRES_TUPLES_OK, query_fmt, args);
    va_end(args);
	return pgres;
}

/* EXPORTED FUNCTIONS */
void
cox_logStart( co_context* cnx )
{
	int id;
    PGresult *pgres;
	request_rec *r;

	r = cnx->request;
	cou_execDBCommand( cnx, "BEGIN");
	cou_execDBCommand( cnx, "
INSERT INTO co_sys_log_request
(id_session, request_num, remote_host, \"user\", request, end_logged) 
VALUES(%d, %d, '%s', '%s', '%s', %d );
",		cou_getSessionId( cnx ), 0, 
		r->connection->remote_ip, 
		cou_buildEncodedSQL(cnx, r->connection->user), 
		/* cou_buildEncodedSQL(cnx, r->the_request), */
		cou_buildEncodedSQL(cnx, r->unparsed_uri),
		1 /* TODO: should be configurable: cou_isEndLogged( cnx ) */
	);
	pgres = cou_execDBSelect( cnx, "
SELECT last_value FROM co_sys_log_request_sqid ;
"	);
	cou_execDBCommand( cnx, "END");
/*
ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME)
r->connection->remote_ip
*/
	if (pgres) { /* got result */
		if (PQntuples(pgres)>0) { /* at least one record returned */
			id = atoi(PQgetvalue(pgres,0,0));
		} else {
			id = 0;
		}
		PQclear(pgres);
	} else {
fprintf( stderr, "Logger: pgres==NULL");
		id = 0;
	}
	/* id = 0 means that something went wrong */
	cou_setLogRequestId( cnx, id );
}

void
cox_logEnd( co_context* cnx )
{
	cou_execDBCommand( cnx, "
UPDATE co_sys_log_request
SET endtime=CURRENT_TIMESTAMP
WHERE id=%d
",	cou_getLogRequestId(cnx) );
/*
TODO:
 state        | integer   | default 0
 size         | integer   | default 0
*/
}

static
const char*
cou_buildEncodedNewLines
(
    co_context *cnx,
    const char *source
)
{
    size_t len=0;
    const char *src_ptr;
    char *dst_ptr;
    char c;
    char *dest;
    /* find length of the destination buffer, */
    /* including escaped newlines */
    src_ptr = source;
    while( (c = *src_ptr++) ){ len++; if (c=='\n') len++; } 
    len++; /* ending NULL */
    dest = (char*)ap_palloc( cnx->request->pool, len );

    dst_ptr = dest;
    while( (c = *source++) ){
        if (c=='\n') { *dst_ptr++='\\'; *dst_ptr++='\n'; }
        else *dst_ptr++ = c;
    }
    *dst_ptr = '\0';
    return dest;
}

/* this is _not_ database function, it uses logger! */
/*
static
void
cou_logMessage( co_context* cnx,
		const char *file, int line, 
        int log_level,
        const char* area_name,
        const char* class_name,
        const char* method_name,
        const char* message
 )
{
	FILE *logf;
	char *errstr;

	if ((log_level & CO_LOG_LEVELMASK) > cnx->request->server->loglevel)
		return;
	errstr = ap_psprintf( cnx->request->pool, 
		"%d|%d|%s|%s|%s|%s[%d] %s",
		cou_getLogRequestId( cnx ),
		log_level,
		area_name,
		class_name,
		method_name,
		file, line,
		cou_buildEncodedNewLines( cnx, message )
	);

	logf = cox_getCOErrorLog( cnx );
    if (logf) {
        fputs(errstr, logf);
        fputc('\n', logf);
        fflush(logf);
    } else {
		ap_log_rerror(file, line, log_level, (cnx)->request, "%s", message );
	}
}
*/
static
void
cou_logErrorCoreV( co_context* cnx, 
	const char *file, int line, 
	int level, const char* fmt, va_list args )
{
    char message[MAX_STRING_LEN];
	FILE *logf;

	if ((level & CO_LOG_LEVELMASK) > cnx->request->server->loglevel)
		return;

	ap_vsnprintf(message, sizeof(message), fmt, args);
/*
see http_log.c
*/
	logf = cox_getCOErrorLog( cnx );
    if (logf) {
        fprintf(logf, "%d|%d|%s|%s|%s|%s[%d] %s\n",
			cou_getLogRequestId( cnx ),
			level,
			cnx->area_name,
			cnx->class_name,
			cnx->method_name,
			file, line,
			cou_buildEncodedNewLines( cnx, message )
		);
        //fputs(errstr, logf);
        //fputc('\n', logf);
        fflush(logf);
    } else {
		ap_log_rerror(file, line, level, (cnx)->request, "%s", message );
	}
}

void
cou_logErrorCore( co_context* cnx, 
	const char *file, int line, 
	int level, const char* fmt, ... )
{
	va_list args;

	va_start(args, fmt);
	cou_logErrorCoreV( cnx, file, line, level, fmt, args );
	va_end(args);

}
