/* to compile - just: gcc logger.c */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

/* open */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* bzero */
#include <string.h>

/* PostgreSQL access functions */
#include "libpq-fe.h"

#include "logger.h"

/* Protocol used:
id_log_request|log_level|area|class|method|message\
message contd.\
message contd.
*/

typedef struct {
	int ready;
	int id_log_request;
	int log_level;
	char area_name[MAX_NAME_SIZE+1];
	char class_name[MAX_NAME_SIZE+1];
	char method_name[MAX_NAME_SIZE+1];
	char message[MAX_MESSAGE_SIZE+1];
	char* message_ptr;
} log_message;

//extern int errno;

static PGconn      *conn = NULL;

/*
 * this function changes all occurences of ' to '', \ to \\
 * but backslash-newline combination is converted to \n
 * args:
 *  destination buffer
 *  length of the destination buffer
 *  source (NULL-ended)
 *  len = number of chars to be copied
 */
static
char*
cou_encodeSQL
(
    char *dest,
    size_t n,
    const char *source,
    size_t len
)
{
    char* limit = dest + n - 1;
    while ((*source) && (dest<limit) && len) { //because we may skip 2 at a time
        if (*source=='\'') { *dest++='\''; *dest='\''; }
        else if (*source=='\\') {
			if ((len>1) && (*(source+1)=='\n')) {
				/* backslash-newline: write now, skip newline in source */
				*dest++='\\'; source++; len--; *dest='n';
			}
			else { *dest++='\\'; *dest='\\'; }
		}
        else *dest=*source;
        dest++; source++; len--;
    }
    //if (dest>limit) { *limit = '\0'; return 1; }
    //if (*--source) { *limit = '\0'; return 1; }
    return dest;
}

void
send_log_record(log_message* rec)
{
    PGresult *res;
	static char query[MAX_QUERY_LEN+1];
	if (!rec->ready) return; 
	/* do only if record is OK */

	//fprintf( stdout, "Sending...\n");

	*(rec->message_ptr) = '\0';
	snprintf(query, MAX_QUERY_LEN, "
INSERT INTO co_sec_log_message 
(id_log_request, log_level, area_name, class_name, method_name, message)
VALUES (%d, %d, '%s', '%s', '%s', '%s')
", 	rec->id_log_request, rec->log_level,
	rec->area_name, rec->class_name, rec->method_name,
	rec->message);

    res = PQexec(conn, query);

    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
		fprintf( stderr, 
			"Error: SQL:%s\nError: for id_log_request=%d\n", 
			PQresultErrorMessage(res), rec->id_log_request
		);
		PQclear(res);
		return;
    }

/*
	fprintf( stdout, "id_log_request: %d\n", rec->id_log_request);
	fprintf( stdout, "log_level: %d\n", rec->log_level);
	fprintf( stdout, "area_name: %s\n", rec->area_name);
	fprintf( stdout, "class_name: %s\n", rec->class_name);
	fprintf( stdout, "method_name: %s\n", rec->method_name);
	fprintf( stdout, "Message:\n%s", rec->message);
*/
	PQclear(res);
	rec->ready = 0; /* never send the same record twice */

}

log_message*
append_log_message(log_message* rec, const char* in, size_t len)
{
	char *dest_after;
	/* check for overflow */
	if (rec->message_ptr+len>rec->message+MAX_MESSAGE_SIZE) {
		/* message too long - buffer too small */
		rec->ready = 0;
		//XXX TODO !!!
		fprintf( stderr, "Error: Overflow\n");
		exit(EXIT_FAILURE);
	}

	/* append message fragment */
	dest_after = cou_encodeSQL(rec->message_ptr, MAX_MESSAGE_SIZE, in, len);
	rec->message_ptr = dest_after;
	return NULL;
}

static log_message curr_record;

/*
 * parses int (written as max MAX_INT_SIZE char + CHAR_SEPARATOR)
 * sets input buffer ptr past the CHAR_SEPARATOR
 * on error returns 0
 * if(res = parse_int(&in)) {do something}
 */
int
parse_int(const char**in_addr)
{
	static char number[MAX_INT_SIZE+1];
	char *end_ptr;

	end_ptr = memccpy( number, *in_addr, CHAR_SEPARATOR, MAX_INT_SIZE+1);
	if (!end_ptr) return 0;
	*(end_ptr-1) = '\0';
	*in_addr += end_ptr - number;
	return atoi(number);
}

char*
parse_name(char* buf, const char**in_addr)
{
	char *end_ptr;

	end_ptr = memccpy( buf, *in_addr, CHAR_SEPARATOR, MAX_NAME_SIZE+1);
	if (!end_ptr) {
		fprintf( stderr, "Error: parse_name %s\n", *in_addr);
		return NULL;
	}
	*(end_ptr-1) = '\0';
	*in_addr += end_ptr - buf;
	return end_ptr;
}

/* 
 * in - string buffer containing beginning of log string
 * in has to be null-ending end has to fit in rec->message buffer
 * ie:
id_log_request|log_level|area|class|method|messagebegining
 * returns NULL if in does not contain correct
*/
log_message*
parse_log_record(log_message* rec, const char* in, size_t len)
{
	static char number[MAX_INT_SIZE+1];
	const char *in_ptr;
	/* init message */
	rec->message_ptr = rec->message;
	rec->ready = 0;

	/* remember beginning */
	in_ptr = in;

	/* parse fields */
	rec->id_log_request=parse_int(&in);
	if (!rec->id_log_request) return NULL;

	rec->log_level=parse_int(&in);
	if (!rec->log_level) return NULL;

	if (!parse_name(rec->area_name, &in)) return NULL;
	if (!parse_name(rec->class_name, &in)) return NULL;
	if (!parse_name(rec->method_name, &in)) return NULL;

	/* append message fragment */
	/* set in, len to correct values */
	rec->ready = 1;
	return append_log_message(rec, in, len - (in-in_ptr));
	// rec on success, NULL on failure
}


/* connect to the database at startup */
static
void 
connect_log()
{
//	if (conn == NULL) {
		conn = PQconnectdb(DB_CONNECTION_STRING);
//	};

	/* check to see that the backend connection was successfully made */
	if (PQstatus(conn) == CONNECTION_BAD)
	{
		fprintf( stderr, "Error: SQL: Connection failed\nError: SQL: %s\n", PQerrorMessage(conn));
		PQfinish(conn);
		exit(EXIT_FAILURE);
	}
}

static
void 
disconnect_log()
{
	PQfinish(conn);
}

main()
{
	int fifo_fd;
	char buf[BUFFER_SIZE];
	char *buf_ptr;
	char *buf_last;
	ssize_t read_len;
	int begin_new_line;
	int send_order; /* send log message to the database */
	int cont_order; /* continue receiving log message */
	int start_order; /* init log message */

	errno = 0;
	bzero(buf, BUFFER_SIZE);

	mkfifo(LOGGER_FIFO, FIFO_MODE);
	fifo_fd = open(LOGGER_FIFO, O_RDONLY);
	if (fifo_fd == -1) {
		fprintf( stderr, "Error: Cannot open FIFO\n");
		exit(EXIT_FAILURE);
	}

	errno = 0;
	connect_log();
	begin_new_line = 1; /* expect order, not line continuation */
	do {
		send_order = 0;
		cont_order = 0;
		start_order = 0;
		read_len = read( fifo_fd, &buf, sizeof(buf)-1);
		if (read_len>0) {
			buf_last = buf;
			for (buf_ptr = buf; buf_ptr<buf+read_len; buf_ptr++) {
				if ((*buf_ptr == '\n')&&(buf_ptr>buf)&&(*(buf_ptr-1)!='\\')) {
					/* send last message, start new message */
					parse_log_record(&curr_record, buf_last, buf_ptr-buf_last);
					send_log_record(&curr_record);
					buf_last = buf_ptr+1;
				}
			}
		}
	} while (1);

	/* TODO: signal handling */
	fprintf( stdout, "Exiting\n");
	disconnect_log();
	close(fifo_fd);
	unlink(LOGGER_FIFO);
	exit(EXIT_SUCCESS);
}